# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from oslo_config import cfg
from oslo_log import log as logging

from neutron.plugins.ml2.drivers.mlnx.agent import config  # noqa

LOG = logging.getLogger(__name__)


class RetryDecorator(object):
    """Retry decorator reruns a method 'retries' times if an exception occurs.

    Decorator for retrying a method if exceptionToCheck exception occurs
    If method raises exception, retries 'retries' times with increasing
    back off period between calls with 'interval' multiplier

    :param exceptionToCheck: the exception to check
    :param interval: initial delay between retries in seconds
    :param retries: number of times to try before giving up
    :raises: exceptionToCheck
    """

    def __init__(self, exceptionToCheck,
                 interval=cfg.CONF.ESWITCH.request_timeout / 1000,
                 retries=cfg.CONF.ESWITCH.retries,
                 backoff_rate=cfg.CONF.ESWITCH.backoff_rate):
        self.exc = exceptionToCheck
        self.interval = interval
        self.retries = retries
        self.backoff_rate = backoff_rate

    def __call__(self, original_func):
        def decorated(*args, **kwargs):
            sleep_interval = self.interval
            num_of_iter = self.retries
            while num_of_iter > 0:
                try:
                    return original_func(*args, **kwargs)
                except self.exc:
                    LOG.debug("Request timeout - call again after "
                              "%s seconds", sleep_interval)
                    time.sleep(sleep_interval)
                    num_of_iter -= 1
                    sleep_interval *= self.backoff_rate

            return original_func(*args, **kwargs)
        return decorated



# Copyright 2015 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import requests

from neutron.common import constants as neutron_const
from neutron.plugins.common import constants
from neutron.plugins.ml2 import config as config
from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2 import plugin
from neutron.tests import base
from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron.tests.unit import testlib_api
from oslo_serialization import jsonutils

from networking_mlnx.plugins.ml2.drivers.sdn import constants as sdn_const
from networking_mlnx.plugins.ml2.drivers.sdn import sdn_mech_driver

PLUGIN_NAME = 'neutron.plugins.ml2.plugin.Ml2Plugin'
SEG_ID = 4L
DEVICE_OWNER_COMPUTE = 'compute:None'
MECHANISM_DRIVER_NAME = 'sdnmechdriver'


class SDNTestCase(test_plugin.Ml2PluginV2TestCase):
    _mechanism_drivers = ['logger', MECHANISM_DRIVER_NAME]

    def setUp(self):
        config.cfg.CONF.set_override('url', 'http://127.0.0.1/neo/cloudx',
                                     sdn_const.GROUP_OPT)
        config.cfg.CONF.set_override('username', 'admin', sdn_const.GROUP_OPT)
        config.cfg.CONF.set_override('password', 'admin', sdn_const.GROUP_OPT)

        super(SDNTestCase, self).setUp()
        self.mech = sdn_mech_driver.SDNMechanismDriver()
        sdn_mech_driver.SDNMechanismDriver._send_http_request = (
            self.check_send_http_request)

    def check_send_http_request(self, urlpath, data, method):
        self.assertFalse(urlpath.startswith("http://"))


class SDNMechanismConfigTests(testlib_api.SqlTestCase):

    def _set_config(self, url='http://127.0.0.1/neo/cloudx',
                    username='admin',
                    password='admin'):
        config.cfg.CONF.set_override('mechanism_drivers',
                                     ['logger', MECHANISM_DRIVER_NAME],
                                     'ml2')
        config.cfg.CONF.set_override('url', url, sdn_const.GROUP_OPT)
        config.cfg.CONF.set_override('username', username, sdn_const.GROUP_OPT)
        config.cfg.CONF.set_override('password', password, sdn_const.GROUP_OPT)

    def _test_missing_config(self, **kwargs):
        self._set_config(**kwargs)
        self.assertRaises(config.cfg.RequiredOptError,
                          plugin.Ml2Plugin)

    def test_valid_config(self):
        self._set_config()
        plugin.Ml2Plugin()

    def test_missing_url_raises_exception(self):
        self._test_missing_config(url=None)

    def test_missing_username_raises_exception(self):
        self._test_missing_config(username=None)

    def test_missing_password_raises_exception(self):
        self._test_missing_config(password=None)


class DataMatcher(object):

    def __init__(self, context, object_type):
        self._data = context.__dict__["_" + object_type.lower()]
        self._data = jsonutils.dumps(self._data, indent=2)

    def __eq__(self, data):
        return data == self._data

    def __repr__(self):
        return self._data


class SDNDriverTestCase(base.BaseTestCase):

    def setUp(self):
        super(SDNDriverTestCase, self).setUp()
        config.cfg.CONF.set_override('mechanism_drivers',
                                     ['logger', MECHANISM_DRIVER_NAME], 'ml2')
        config.cfg.CONF.set_override('url', 'http://127.0.0.1/neo/cloudx',
                                     sdn_const.GROUP_OPT)
        config.cfg.CONF.set_override('username', 'admin', sdn_const.GROUP_OPT)
        config.cfg.CONF.set_override('password', 'admin', sdn_const.GROUP_OPT)
        self.mech = sdn_mech_driver.SDNMechanismDriver()
        self.mech.initialize()

    def _get_segments_list(self, seg_id=SEG_ID, net_type=constants.TYPE_VLAN):
        return [{'segmentation_id': seg_id,
                'physical_network': u'physnet1',
                'id': u'72770a8a-e9b4-46da-8f0a-ffbd6a7fa3de',
                'network_type': net_type}]

    def _get_mock_network_operation_context(self):
        current = {"provider:segmentation_id": SEG_ID,
                   'id': 'd897e21a-dfd6-4331-a5dd-7524fa421c3e',
                   'name': 'net1',
                   'provider:network_type': 'vlan',
                   'network_qos_policy': None}
        context = mock.Mock(current=current, _network=current,
                            _segments=self._get_segments_list())
        return context

    def _get_mock_port_operation_context(self):
        current = {'binding:host_id': 'r-ufm177',
                   'binding:profile': {u'pci_slot': u'0000:02:00.4',
                                       u'physical_network': u'physnet1',
                                       u'pci_vendor_info': u'15b3:1004'},
                   'id': '72c56c48-e9b8-4dcf-b3a7-0813bb3bd839',
                   'binding:vnic_type': 'direct',
                   'mac_address': '12:34:56:78:21:b6',
                   'name': 'port_test1',
                   'network_id': 'c13bba05-eb07-45ba-ace2-765706b2d701',
                   'network_qos_policy': None}

        # The port context should have NetwrokContext object that contain
        # the segments list
        network_context = type('NetworkContext', (object,),
                            {"_segments": self._get_segments_list()})

        context = mock.Mock(current=current, _port=current,
                            _network_context=network_context)
        return context

    def _get_mock_bind_operation_context(self,
                                         device_owner=DEVICE_OWNER_COMPUTE):
        current = {'device_owner': device_owner,
                   'network_id': 'c13bba05-eb07-45ba-ace2-765706b2d701',
                   'network_qos_policy': None}
        context = mock.Mock(current=current, _port=current,
                            segments_to_bind=self._get_segments_list())
        return context

    def _get_mock_operation_context(self, object_type):
        getter = getattr(self, '_get_mock_%s_operation_context' %
                         object_type.lower())
        return getter()

    _status_code_msgs = {
        200: '',
        201: '',
        204: '',
        400: '400 Client Error: Bad Request',
        401: '401 Client Error: Unauthorized',
        403: '403 Client Error: Forbidden',
        404: '404 Client Error: Not Found',
        409: '409 Client Error: Conflict',
        501: '501 Server Error: Not Implemented',
        503: '503 Server Error: Service Unavailable',
    }

    @classmethod
    def _get_mock_request_response(cls, status_code):
        response = mock.Mock(status_code=status_code)
        response.raise_for_status = mock.Mock() if status_code < 400 else (
            mock.Mock(side_effect=requests.exceptions.HTTPError(
                cls._status_code_msgs[status_code])))
        return response

    def _get_http_request_codes(self):
        for err_code in (requests.codes.ok,
                         requests.codes.created,
                         requests.codes.no_content,
                         requests.codes.bad_request,
                         requests.codes.unauthorized,
                         requests.codes.forbidden,
                         requests.codes.not_found,
                         requests.codes.conflict,
                         requests.codes.not_implemented,
                         requests.codes.service_unavailable):
            yield err_code

    def _test_no_operation(self, method, context, status_code,
                           *args, **kwargs):
        request_response = self._get_mock_request_response(status_code)
        with mock.patch('requests.Session.request',
                    return_value=request_response) as mock_method:
            method(context)
            assert not mock_method.called, ('Expected not to be called. '
                                       'Called %d times' % mock_method.calls)

    def _test_operation_with(self, method, context, status_code,
                           *args, **kwargs):
        request_response = self._get_mock_request_response(status_code)
        with mock.patch('requests.Session.request',
                        return_value=request_response) as mock_method:
            method(context)
            mock_method.assert_called_with(
                    headers=sdn_const.JSON_HTTP_HEADER,
                    timeout=config.cfg.CONF.sdn.timeout,
                    *args, **kwargs)

    def _test_create_resource_postcommit(self, object_type, status_code):
        method = getattr(self.mech, 'create_%s_postcommit' %
                         object_type.lower())
        context = self._get_mock_operation_context(object_type)
        url = '%s/%s' % (config.cfg.CONF.sdn.url, object_type)
        kwargs = {'url': url, 'data': DataMatcher(context, object_type)}
        self._test_operation_with(method, context, status_code,
                           sdn_const.POST, **kwargs)

    def _test_update_resource_postcommit(self, object_type, status_code):
        method = getattr(self.mech, 'update_%s_postcommit' %
                         object_type.lower())
        context = self._get_mock_operation_context(object_type)
        url = '%s/%s/%s' % (config.cfg.CONF.sdn.url, object_type,
                            context.current['id'])
        kwargs = {'url': url, 'data': DataMatcher(context, object_type)}
        self._test_operation_with(method, context, status_code,
                           sdn_const.PUT, **kwargs)

    def _test_delete_resource_postcommit(self, object_type, status_code):
        method = getattr(self.mech, 'delete_%s_postcommit' %
                         object_type.lower())
        context = self._get_mock_operation_context(object_type)
        url = '%s/%s/%s' % (config.cfg.CONF.sdn.url, object_type,
                            context.current['id'])
        kwargs = {'url': url, 'data': DataMatcher(context, object_type)}
        self._test_operation_with(method, context, status_code,
                           sdn_const.DELETE, **kwargs)

    def _test_bind_port(self, status_code, context, assert_called=True):
        method = getattr(self.mech, 'bind_port')
        object_type = sdn_const.PORT_PATH
        url = '%s/%s' % (config.cfg.CONF.sdn.url, object_type)
        kwargs = {'url': url, 'data': DataMatcher(context, object_type)}

        if assert_called:
            self._test_operation_with(method, context, status_code,
                                    sdn_const.POST, **kwargs)
        else:
            self._test_no_operation(method, context, status_code,
                                    sdn_const.POST, **kwargs)

    @mock.patch('neutron.objects.qos.policy.QosPolicy.get_network_policy',
                return_value=None)
    def test_create_network_postcommit(self, *args):
        for status_code in self._get_http_request_codes():
            self._test_create_resource_postcommit(sdn_const.NETWORK_PATH,
                                                  status_code,
                                                  )

    @mock.patch('neutron.objects.qos.policy.QosPolicy.get_network_policy',
                return_value=None)
    def test_update_port_postcommit(self, *args):
        for status_code in self._get_http_request_codes():
            self._test_update_resource_postcommit(sdn_const.PORT_PATH,
                                                  status_code,
                                                  )

    @mock.patch('neutron.objects.qos.policy.QosPolicy.get_network_policy',
                return_value=None)
    def test_update_network_postcommit(self, *args):
        for status_code in self._get_http_request_codes():
            self._test_update_resource_postcommit(sdn_const.NETWORK_PATH,
                                                  status_code,
                                                  )

    @mock.patch('neutron.objects.qos.policy.QosPolicy.get_network_policy',
                return_value=None)
    def test_delete_network_postcommit(self, *args):
        for status_code in self._get_http_request_codes():
            self._test_delete_resource_postcommit(sdn_const.NETWORK_PATH,
                                                  status_code,
                                                  )

    @mock.patch('neutron.objects.qos.policy.QosPolicy.get_network_policy',
                return_value=None)
    def test_delete_port_postcommit(self, *args):
        for status_code in self._get_http_request_codes():
            self._test_delete_resource_postcommit(sdn_const.PORT_PATH,
                                                  status_code,
                                                  )

    @mock.patch('neutron.objects.qos.policy.QosPolicy.get_network_policy',
                return_value=None)
    def test_bind_port_compute(self, *args):
        """Bind port to VM

        SDN MD will call this kind of bind only
        The identify of this call is in port context
        The device_owner should be: "compute:None"
        """
        context = self._get_mock_bind_operation_context()
        for status_code in self._get_http_request_codes():
            self._test_bind_port(status_code, context)

    @mock.patch('neutron.objects.qos.policy.QosPolicy.get_network_policy',
                return_value=None)
    def test_bind_port_dhcp(self, *args):
        """Bind port dhcp context

        bind network port occurs when a port binded to a dhcp
        SDN MD will filter such a calls
        """
        context = self._get_mock_bind_operation_context(
                                        neutron_const.DEVICE_OWNER_DHCP)
        for status_code in self._get_http_request_codes():
            self._test_bind_port(status_code, context)

    @mock.patch('neutron.objects.qos.policy.QosPolicy.get_network_policy',
                return_value=None)
    def test_bind_port_router(self, *args):
        """Bind port router context

        bind network port occurs when a port binded to a router
        SDN MD will filter such a calls
        """
        context = self._get_mock_bind_operation_context(
                                        neutron_const.DEVICE_OWNER_ROUTER_GW)
        for status_code in self._get_http_request_codes():
            self._test_bind_port(status_code, context, assert_called=False)

    def test_check_segment(self):
        """Validate the check_segment call."""
        segment = {'api.NETWORK_TYPE': ""}
        segment[api.NETWORK_TYPE] = constants.TYPE_VLAN
        self.assertTrue(self.mech.check_segment(segment))
        # Validate a network type not currently supported
        segment[api.NETWORK_TYPE] = constants.TYPE_LOCAL
        self.assertFalse(self.mech.check_segment(segment))
        segment[api.NETWORK_TYPE] = constants.TYPE_FLAT
        self.assertFalse(self.mech.check_segment(segment))
        segment[api.NETWORK_TYPE] = constants.TYPE_GRE
        self.assertFalse(self.mech.check_segment(segment))
        segment[api.NETWORK_TYPE] = constants.TYPE_VXLAN
        self.assertFalse(self.mech.check_segment(segment))



# Copyright 2015 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from neutron.tests.unit.agent.linux import test_dhcp

from networking_mlnx.dhcp import mlnx_dhcp


class TestMlnxDnsmasq(test_dhcp.TestDnsmasq):

    def _get_dnsmasq(self, network, process_monitor=None):
        process_monitor = process_monitor or mock.Mock()
        return mlnx_dhcp.MlnxDnsmasq(self.conf, network,
                            process_monitor=process_monitor)

    def test_only_populates_dhcp_enabled_subnet_on_a_network(self):
        exp_host_name = '/dhcp/cccccccc-cccc-cccc-cccc-cccccccccccc/host'
        exp_host_data = ('00:00:80:aa:bb:cc,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:80:00:'
                         '00:aa:bb:cc,host-192-168-0-2.openstacklocal.,'
                         '192.168.0.2\n'
                         '00:00:f3:aa:bb:cc,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:f3:00:'
                         '00:aa:bb:cc,host-192-168-0-3.openstacklocal.,'
                         '192.168.0.3\n'
                         '00:00:0f:aa:bb:cc,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:0f:00:'
                         '00:aa:bb:cc,host-192-168-0-4.openstacklocal.,'
                         '192.168.0.4\n'
                         '00:00:0f:rr:rr:rr,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:0f:00:'
                         '00:rr:rr:rr,host-192-168-0-1.openstacklocal.,'
                         '192.168.0.1\n').lstrip()
        dm = self._get_dnsmasq(test_dhcp.FakeDualNetworkSingleDHCP())
        dm._output_hosts_file()
        self.safe.assert_has_calls([mock.call(exp_host_name,
                                              exp_host_data)])

    @property
    def _test_no_dhcp_domain_alloc_data(self):
        exp_host_name = '/dhcp/cccccccc-cccc-cccc-cccc-cccccccccccc/host'
        exp_host_data = ('00:00:80:aa:bb:cc,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:80:00:'
                         '00:aa:bb:cc,host-192-168-0-2,'
                         '192.168.0.2\n'
                         '00:00:f3:aa:bb:cc,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:f3:00:'
                         '00:aa:bb:cc,host-fdca-3ba5-a17a-4ba3--2,'
                         '[fdca:3ba5:a17a:4ba3::2]\n'
                         '00:00:0f:aa:bb:cc,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:0f:00:'
                         '00:aa:bb:cc,host-192-168-0-3,'
                         '192.168.0.3\n'
                         '00:00:0f:aa:bb:cc,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:0f:00:'
                         '00:aa:bb:cc,host-fdca-3ba5-a17a-4ba3--3,'
                         '[fdca:3ba5:a17a:4ba3::3]\n'
                         '00:00:0f:rr:rr:rr,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:0f:00:'
                         '00:rr:rr:rr,host-192-168-0-1,'
                         '192.168.0.1\n').lstrip()
        exp_addn_name = '/dhcp/cccccccc-cccc-cccc-cccc-cccccccccccc/addn_hosts'
        exp_addn_data = (
            '192.168.0.2\t'
            'host-192-168-0-2 host-192-168-0-2\n'
            'fdca:3ba5:a17a:4ba3::2\t'
            'host-fdca-3ba5-a17a-4ba3--2 '
            'host-fdca-3ba5-a17a-4ba3--2\n'
            '192.168.0.3\thost-192-168-0-3 '
            'host-192-168-0-3\n'
            'fdca:3ba5:a17a:4ba3::3\t'
            'host-fdca-3ba5-a17a-4ba3--3 '
            'host-fdca-3ba5-a17a-4ba3--3\n'
            '192.168.0.1\t'
            'host-192-168-0-1 '
            'host-192-168-0-1\n'
        ).lstrip()
        return (exp_host_name, exp_host_data,
                exp_addn_name, exp_addn_data)

    @property
    def _test_reload_allocation_data(self):
        exp_host_name = '/dhcp/cccccccc-cccc-cccc-cccc-cccccccccccc/host'
        exp_host_data = ('00:00:80:aa:bb:cc,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:80:00:'
                         '00:aa:bb:cc,host-192-168-0-2.openstacklocal.,'
                         '192.168.0.2\n'
                         '00:00:f3:aa:bb:cc,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:f3:00:'
                         '00:aa:bb:cc,host-fdca-3ba5-a17a-4ba3--2.'
                         'openstacklocal.,[fdca:3ba5:a17a:4ba3::2]\n'
                         '00:00:0f:aa:bb:cc,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:0f:00:'
                         '00:aa:bb:cc,host-192-168-0-3.openstacklocal.,'
                         '192.168.0.3\n'
                         '00:00:0f:aa:bb:cc,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:0f:00:'
                         '00:aa:bb:cc,host-fdca-3ba5-a17a-4ba3--3.'
                         'openstacklocal.,[fdca:3ba5:a17a:4ba3::3]\n'
                         '00:00:0f:rr:rr:rr,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:0f:00:'
                         '00:rr:rr:rr,host-192-168-0-1.openstacklocal.,'
                         '192.168.0.1\n').lstrip()
        exp_addn_name = '/dhcp/cccccccc-cccc-cccc-cccc-cccccccccccc/addn_hosts'
        exp_addn_data = (
            '192.168.0.2\t'
            'host-192-168-0-2.openstacklocal. host-192-168-0-2\n'
            'fdca:3ba5:a17a:4ba3::2\t'
            'host-fdca-3ba5-a17a-4ba3--2.openstacklocal. '
            'host-fdca-3ba5-a17a-4ba3--2\n'
            '192.168.0.3\thost-192-168-0-3.openstacklocal. '
            'host-192-168-0-3\n'
            'fdca:3ba5:a17a:4ba3::3\t'
            'host-fdca-3ba5-a17a-4ba3--3.openstacklocal. '
            'host-fdca-3ba5-a17a-4ba3--3\n'
            '192.168.0.1\t'
            'host-192-168-0-1.openstacklocal. '
            'host-192-168-0-1\n'
        ).lstrip()
        exp_opt_name = '/dhcp/cccccccc-cccc-cccc-cccc-cccccccccccc/opts'
        fake_v6 = '2001:0200:feed:7ac0::1'
        exp_opt_data = (
            'tag:tag0,option:dns-server,8.8.8.8\n'
            'tag:tag0,option:classless-static-route,20.0.0.1/24,20.0.0.1,'
            '169.254.169.254/32,192.168.0.1,0.0.0.0/0,192.168.0.1\n'
            'tag:tag0,249,20.0.0.1/24,20.0.0.1,'
            '169.254.169.254/32,192.168.0.1,0.0.0.0/0,192.168.0.1\n'
            'tag:tag0,option:router,192.168.0.1\n'
            'tag:tag1,option6:dns-server,%s\n'
            'tag:tag1,option6:domain-search,openstacklocal').lstrip() % (
            '[' + fake_v6 + ']')
        return (exp_host_name, exp_host_data,
                exp_addn_name, exp_addn_data,
                exp_opt_name, exp_opt_data,)

    def test_release_unused_leases_one_lease_with_client_id(self):
        dnsmasq = self._get_dnsmasq(test_dhcp.FakeDualNetwork())

        ip1 = '192.168.0.2'
        mac1 = '00:00:80:aa:bb:cc'
        client_id1 = ('ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:80:'
                      '00:00:aa:bb:cc')
        ip2 = '192.168.0.5'
        mac2 = '00:00:0f:aa:bb:55'
        client_id2 = ('ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:0f:'
                      '00:00:aa:bb:55')

        old_leases = set([(ip1, mac1, client_id1), (ip2, mac2, client_id2)])
        dnsmasq._read_hosts_file_leases = mock.Mock(return_value=old_leases)
        dnsmasq._output_hosts_file = mock.Mock()
        dnsmasq._release_lease = mock.Mock()
        dnsmasq.network.ports = [test_dhcp.FakePort5()]

        dnsmasq._release_unused_leases()

        dnsmasq._release_lease.assert_called_once_with(
            mac1, ip1, client_id1)

    def test_only_populates_dhcp_client_id(self):
        exp_host_name = '/dhcp/aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/host'
        exp_host_data = ('00:00:80:aa:bb:cc,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:80:'
                         '00:00:aa:bb:cc,host-192-168-0-2.openstacklocal.,'
                         '192.168.0.2\n'
                         '00:00:0f:aa:bb:55,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:0f:'
                         '00:00:aa:bb:55,host-192-168-0-5.openstacklocal.,'
                         '192.168.0.5\n'
                         '00:00:0f:aa:bb:66,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:0f:'
                         '00:00:aa:bb:66,host-192-168-0-6.openstacklocal.,'
                         '192.168.0.6,'
                         'set:ccccccccc-cccc-cccc-cccc-ccccccccc\n').lstrip()

        dm = self._get_dnsmasq(test_dhcp.FakeV4NetworkClientId)
        dm._output_hosts_file()
        self.safe.assert_has_calls([mock.call(exp_host_name,
                                              exp_host_data)])

    def test_host_and_opts_file_on_net_with_V6_stateless_and_V4_subnets(
                                                                    self):
        exp_host_name = '/dhcp/bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb/host'
        exp_host_data = (
            '00:16:3e:c2:77:1d,set:hhhhhhhh-hhhh-hhhh-hhhh-hhhhhhhhhhhh\n'
            '00:16:3e:c2:77:1d,'
            'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:16:3e:00:00:c2:77:1d,'
            'host-192-168-0-3.openstacklocal.,'
            '192.168.0.3,set:hhhhhhhh-hhhh-hhhh-hhhh-hhhhhhhhhhhh\n'
            '00:00:0f:rr:rr:rr,'
            'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:0f:00:00:rr:rr:rr,'
            'host-192-168-0-1.openstacklocal.,192.168.0.1\n').lstrip()
        exp_opt_name = '/dhcp/bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb/opts'
        exp_opt_data = (
            'tag:tag0,option6:domain-search,openstacklocal\n'
            'tag:tag1,option:dns-server,8.8.8.8\n'
            'tag:tag1,option:classless-static-route,20.0.0.1/24,20.0.0.1,'
            '169.254.169.254/32,192.168.0.1,0.0.0.0/0,192.168.0.1\n'
            'tag:tag1,249,20.0.0.1/24,20.0.0.1,169.254.169.254/32,'
            '192.168.0.1,0.0.0.0/0,192.168.0.1\n'
            'tag:tag1,option:router,192.168.0.1\n'
            'tag:hhhhhhhh-hhhh-hhhh-hhhh-hhhhhhhhhhhh,'
            'option6:dns-server,ffea:3ba5:a17a:4ba3::100').lstrip()

        dm = self._get_dnsmasq(
            test_dhcp.FakeNetworkWithV6SatelessAndV4DHCPSubnets())
        dm._output_hosts_file()
        dm._output_opts_file()
        self.safe.assert_has_calls([mock.call(exp_host_name, exp_host_data),
                                    mock.call(exp_opt_name, exp_opt_data)])

    def test_only_populates_dhcp_enabled_subnets(self):
        exp_host_name = '/dhcp/eeeeeeee-eeee-eeee-eeee-eeeeeeeeeeee/host'
        exp_host_data = ('00:00:80:aa:bb:cc,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:80:'
                         '00:00:aa:bb:cc,host-192-168-0-2.openstacklocal.,'
                         '192.168.0.2\n'
                         '00:16:3E:C2:77:1D,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:16:3E:'
                         '00:00:C2:77:1D,host-192-168-0-4.openstacklocal.,'
                         '192.168.0.4\n'
                         '00:00:0f:rr:rr:rr,'
                         'id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:0f:'
                         '00:00:rr:rr:rr,host-192-168-0-1.openstacklocal.,'
                         '192.168.0.1\n').lstrip()
        dm = self._get_dnsmasq(test_dhcp.FakeDualStackNetworkSingleDHCP())
        dm._output_hosts_file()
        self.safe.assert_has_calls([mock.call(exp_host_name,
                                              exp_host_data)])

    def test_release_unused_leases_one_lease(self):
        pass

    def test_host_file_on_net_with_v6_slaac_and_v4(self):
        exp_host_name = '/dhcp/eeeeeeee-eeee-eeee-eeee-eeeeeeeeeeee/host'
        exp_host_data = (
            '00:00:80:aa:bb:cc,id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:'
            '00:80:00:00:aa:bb:cc,host-192-168-0-2.openstacklocal.,'
            '192.168.0.2,set:eeeeeeee-eeee-eeee-eeee-eeeeeeeeeeee\n'
            '00:16:3E:C2:77:1D,id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:'
            '16:3E:00:00:C2:77:1D,host-192-168-0-4.openstacklocal.,'
            '192.168.0.4,set:gggggggg-gggg-gggg-gggg-gggggggggggg\n'
            '00:00:0f:rr:rr:rr,id:ff:00:00:00:00:00:02:00:00:02:c9:00:00:00:'
            '0f:00:00:rr:rr:rr,host-192-168-0-1.openstacklocal.,192.168.0.1,'
            'set:rrrrrrrr-rrrr-rrrr-rrrr-rrrrrrrrrrrr\n').lstrip()
        dm = self._get_dnsmasq(
            test_dhcp.FakeDualStackNetworkingSingleDHCPTags())
        dm._output_hosts_file()
        self.safe.assert_has_calls([mock.call(exp_host_name, exp_host_data)])



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from oslo_config import cfg

from networking_mlnx.plugins.ml2.drivers.mlnx.agent import comm_utils
from networking_mlnx.plugins.ml2.drivers.mlnx.agent import exceptions
from neutron.plugins.ml2.drivers.mlnx.agent import config  # noqa
from neutron.tests import base


class WrongException(Exception):
        pass


class TestRetryDecorator(base.BaseTestCase):
    def setUp(self):
        super(TestRetryDecorator, self).setUp()
        self.sleep_fn_p = mock.patch("time.sleep")
        self.sleep_fn = self.sleep_fn_p.start()

    def test_no_retry_required(self):
        self.counter = 0

        @comm_utils.RetryDecorator(exceptions.RequestTimeout, interval=2,
                                   retries=3, backoff_rate=2)
        def succeeds():
            self.counter += 1
            return 'success'

        ret = succeeds()
        self.assertFalse(self.sleep_fn.called)
        self.assertEqual(ret, 'success')
        self.assertEqual(self.counter, 1)

    def test_retry_zero_times(self):
        self.counter = 0
        interval = 2
        backoff_rate = 2
        retries = 0

        @comm_utils.RetryDecorator(exceptions.RequestTimeout, interval,
                                   retries, backoff_rate)
        def always_fails():
            self.counter += 1
            raise exceptions.RequestTimeout()

        self.assertRaises(exceptions.RequestTimeout, always_fails)
        self.assertEqual(self.counter, 1)
        self.assertFalse(self.sleep_fn.called)

    def test_retries_once(self):
        self.counter = 0
        interval = 2
        backoff_rate = 2
        retries = 3

        @comm_utils.RetryDecorator(exceptions.RequestTimeout, interval,
                                   retries, backoff_rate)
        def fails_once():
            self.counter += 1
            if self.counter < 2:
                raise exceptions.RequestTimeout()
            else:
                return 'success'

        ret = fails_once()
        self.assertEqual(ret, 'success')
        self.assertEqual(self.counter, 2)
        self.assertEqual(self.sleep_fn.call_count, 1)
        self.sleep_fn.assert_called_with(interval)

    def test_limit_is_reached(self):
        self.counter = 0
        retries = 3
        interval = 2
        backoff_rate = 4

        @comm_utils.RetryDecorator(exceptions.RequestTimeout, interval,
                                   retries, backoff_rate)
        def always_fails():
            self.counter += 1
            raise exceptions.RequestTimeout()

        self.assertRaises(exceptions.RequestTimeout, always_fails)
        self.assertEqual(self.counter, retries + 1)
        self.assertEqual(self.sleep_fn.call_count, retries)

        expected_sleep_fn_arg = []
        for i in range(retries):
            expected_sleep_fn_arg.append(interval)
            interval *= backoff_rate

        self.sleep_fn.assert_has_calls(map(mock.call, expected_sleep_fn_arg))

    def test_limit_is_reached_with_conf(self):
        self.counter = 0

        @comm_utils.RetryDecorator(exceptions.RequestTimeout)
        def always_fails():
            self.counter += 1
            raise exceptions.RequestTimeout()

        retry = cfg.CONF.ESWITCH.retries
        interval = cfg.CONF.ESWITCH.request_timeout / 1000
        delay_rate = cfg.CONF.ESWITCH.backoff_rate

        expected_sleep_fn_arg = []
        for i in range(retry):
            expected_sleep_fn_arg.append(interval)
            interval *= delay_rate

        self.assertRaises(exceptions.RequestTimeout, always_fails)
        self.assertEqual(self.counter, retry + 1)
        self.assertEqual(self.sleep_fn.call_count, retry)
        self.sleep_fn.assert_has_calls(map(mock.call, expected_sleep_fn_arg))

    def test_wrong_exception_no_retry(self):

        @comm_utils.RetryDecorator(exceptions.RequestTimeout)
        def raise_unexpected_error():
            raise WrongException("wrong exception")

        self.assertRaises(WrongException, raise_unexpected_error)
        self.assertFalse(self.sleep_fn.called)



# Copyright 2015 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Config file groups name
GROUP_OPT = "sdn"

# RESTful API paths:
NETWORK_PATH = "Network"
PORT_PATH = "Port"

# HTTP request methods:
DELETE = "DELETE"
POST = "POST"
PUT = "PUT"

# HTTP headers
LOGIN_HTTP_HEADER = {'content-type': 'application/x-www-form-urlencoded'}
JSON_HTTP_HEADER = {"Content-Type": "application/json"}

# Port device owner
PORT_DEVICE_OWNER_COMPUTE = 'compute:'



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import importutils

from networking_mlnx.plugins.ml2.drivers.mlnx.agent import comm_utils
from networking_mlnx.plugins.ml2.drivers.mlnx.agent import exceptions
from neutron.i18n import _LE

zmq = importutils.try_import('eventlet.green.zmq')

LOG = logging.getLogger(__name__)


class EswitchUtils(object):
    def __init__(self, daemon_endpoint, timeout):
        if not zmq:
            LOG.error(_LE("Failed to import eventlet.green.zmq. "
                          "Won't connect to eSwitchD - exiting..."))
            raise SystemExit(1)
        self.__conn = None
        self.daemon = daemon_endpoint
        self.timeout = timeout

    @property
    def _conn(self):
        if self.__conn is None:
            context = zmq.Context()
            socket = context.socket(zmq.REQ)
            socket.setsockopt(zmq.LINGER, 0)
            socket.connect(self.daemon)
            self.__conn = socket
            self.poller = zmq.Poller()
            self.poller.register(self._conn, zmq.POLLIN)
        return self.__conn

    @comm_utils.RetryDecorator(exceptions.RequestTimeout)
    def send_msg(self, msg):
        self._conn.send(msg)

        socks = dict(self.poller.poll(self.timeout))
        if socks.get(self._conn) == zmq.POLLIN:
            recv_msg = self._conn.recv()
            response = self.parse_response_msg(recv_msg)
            return response
        else:
            self._conn.setsockopt(zmq.LINGER, 0)
            self._conn.close()
            self.poller.unregister(self._conn)
            self.__conn = None
            raise exceptions.RequestTimeout()

    def parse_response_msg(self, recv_msg):
        msg = jsonutils.loads(recv_msg)
        if msg['status'] == 'OK':
            if 'response' in msg:
                return msg.get('response')
            return
        elif msg['status'] == 'FAIL':
            msg_dict = dict(action=msg['action'], reason=msg['reason'])
            error_msg = _LE("Action %(action)s failed: %(reason)s") % msg_dict
        else:
            error_msg = _LE("Unknown operation status %s") % msg['status']
        LOG.error(error_msg)
        raise exceptions.OperationFailed(err_msg=error_msg)

    def get_attached_vnics(self):
        LOG.debug("get_attached_vnics")
        msg = jsonutils.dumps({'action': 'get_vnics', 'fabric': '*'})
        vnics = self.send_msg(msg)
        return vnics

    def set_port_vlan_id(self, physical_network,
                         segmentation_id, port_mac):
        LOG.debug("Set Vlan  %(segmentation_id)s on Port %(port_mac)s "
                  "on Fabric %(physical_network)s",
                  {'port_mac': port_mac,
                   'segmentation_id': segmentation_id,
                   'physical_network': physical_network})
        msg = jsonutils.dumps({'action': 'set_vlan',
                               'fabric': physical_network,
                               'port_mac': port_mac,
                               'vlan': segmentation_id})
        self.send_msg(msg)

    def define_fabric_mappings(self, interface_mapping):
        for fabric, phy_interface in interface_mapping.iteritems():
            LOG.debug("Define Fabric %(fabric)s on interface %(ifc)s",
                      {'fabric': fabric,
                       'ifc': phy_interface})
            msg = jsonutils.dumps({'action': 'define_fabric_mapping',
                                   'fabric': fabric,
                                   'interface': phy_interface})
            self.send_msg(msg)

    def port_up(self, fabric, port_mac):
        LOG.debug("Port Up for %(port_mac)s on fabric %(fabric)s",
                  {'port_mac': port_mac, 'fabric': fabric})
        msg = jsonutils.dumps({'action': 'port_up',
                               'fabric': fabric,
                               'ref_by': 'mac_address',
                               'mac': 'port_mac'})
        self.send_msg(msg)

    def port_down(self, fabric, port_mac):
        LOG.debug("Port Down for %(port_mac)s on fabric %(fabric)s",
                  {'port_mac': port_mac, 'fabric': fabric})
        msg = jsonutils.dumps({'action': 'port_down',
                               'fabric': fabric,
                               'ref_by': 'mac_address',
                               'mac': port_mac})
        self.send_msg(msg)

    def port_release(self, fabric, port_mac):
        LOG.debug("Port Release for %(port_mac)s on fabric %(fabric)s",
                  {'port_mac': port_mac, 'fabric': fabric})
        msg = jsonutils.dumps({'action': 'port_release',
                               'fabric': fabric,
                               'ref_by': 'mac_address',
                               'mac': port_mac})
        self.send_msg(msg)

    def get_eswitch_ports(self, fabric):
        # TODO(irena) - to implement for next phase
        return {}

    def get_eswitch_id(self, fabric):
        # TODO(irena) - to implement for next phase
        return ""



# Copyright 2015 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import functools
import requests
import time

from neutron.common import constants as neutron_const
from neutron.i18n import _LE
from neutron.objects.qos import policy as policy_object
from neutron.plugins.common import constants
from neutron.plugins.ml2 import driver_api as api
from oslo_config import cfg
from oslo_log import log
from oslo_serialization import jsonutils

import networking_mlnx.plugins.ml2.drivers.sdn.constants as sdn_const
import networking_mlnx.plugins.ml2.drivers.sdn.exceptions as sdn_exc

LOG = log.getLogger(__name__)

sdn_opts = [
        cfg.StrOpt('url',
                   help=_("HTTP URL of SDN Provider."),
                   ),
        cfg.StrOpt('domain',
                   help=_("Cloud domain name in SDN provider "
                          "(for example: cloudx)"),
                   ),
        cfg.StrOpt('username',
                   help=_("HTTP username for authentication."),
                   ),
        cfg.StrOpt('password',
                   help=_("HTTP password for authentication."),
                   secret=True
                   ),
        cfg.IntOpt('timeout',
                   help=_("HTTP timeout in seconds."),
                   default=10
                   ),
        cfg.IntOpt('session_timeout',
                   help=_("Login session timeout in seconds."),
                   default=86400
                   ),
]

cfg.CONF.register_opts(sdn_opts, sdn_const.GROUP_OPT)
NETWORK_QOS_POLICY = 'network_qos_policy'


def context_validator(context_type=None):
    def real_decorator(func):
        @functools.wraps(func)
        def wrapper(instance, context, *args, **kwargs):
            if context_type == sdn_const.PORT_PATH:
                # port context contain network_context
                # which include the segments
                segments = getattr(context._network_context, "_segments", None)
            elif context_type == sdn_const.NETWORK_PATH:
                segments = getattr(context, "_segments", None)
            else:
                segments = getattr(context, "segments_to_bind", None)
            if segments and getattr(instance, "check_segments", None):
                if instance.check_segments(segments):
                    return func(instance, context, *args, **kwargs)
        return wrapper
    return real_decorator


def error_handler(func):
    @functools.wraps(func)
    def wrapper(instance, *args, **kwargs):
        try:
            return func(instance, *args, **kwargs)
        except Exception as (e):
            LOG.error(
                    _LE("%(function_name)s %(exception_desc)s"),
                    {'function_name': func.__name__,
                    'exception_desc': str(e)}
            )
    return wrapper


class SDNMechanismDriver(api.MechanismDriver):

    """Mechanism Driver for SDN.

    This driver send notifications to SDN provider.
    The notifications are for port/network changes.
    """

    def _validate_mandatory_params_exist(self):
        mandatory_args = ("url", "username", "password")
        for arg in mandatory_args:
            if not getattr(self, arg):
                raise cfg.RequiredOptError(arg, sdn_const.GROUP_OPT)

    def _strings_to_url(self, *args):
        return "/".join(filter(None, args))

    def _getSession(self):
        if self.session_start == 0 or (time.time() - self.session_start
                                       >= self.session_timeout):
            login_url = self._strings_to_url(str(self.url), "login")
            login_data = "username=%s&password=%s" % (self.username,
                                                      self.password)
            login_headers = sdn_const.LOGIN_HTTP_HEADER
            try:
                session = requests.session()
                LOG.debug("Login to SDN Provider. Login URL %(url)s",
                         {'url': login_url})
                r = session.request(sdn_const.POST, login_url, data=login_data,
                                    headers=login_headers)
                LOG.debug("request status: %d", r.status_code)
                r.raise_for_status()
            except Exception as e:
                raise sdn_exc.SDNLoginError(login_url=login_url, msg=e)

            self.session_start = time.time()
            self.session = session

        return self.session

    def initialize(self):
        self.url = cfg.CONF.sdn.url
        try:
            self.url = self.url.rstrip("/")
        except Exception:
            pass
        self.domain = cfg.CONF.sdn.domain
        self.username = cfg.CONF.sdn.username
        self.password = cfg.CONF.sdn.password
        self.timeout = cfg.CONF.sdn.timeout
        self.session_timeout = cfg.CONF.sdn.session_timeout
        self.session_start = 0
        self._validate_mandatory_params_exist()

    @context_validator(sdn_const.NETWORK_PATH)
    @error_handler
    def create_network_postcommit(self, context):
        network_dic = context._network
        network_dic[NETWORK_QOS_POLICY] = (
            self._get_network_qos_policy(context, network_dic['id']))
        self._send_json_http_request(method=sdn_const.POST,
                                     urlpath=sdn_const.NETWORK_PATH,
                                     data=network_dic)

    @context_validator(sdn_const.NETWORK_PATH)
    @error_handler
    def update_network_postcommit(self, context):
        network_dic = context._network
        network_dic[NETWORK_QOS_POLICY] = (
            self._get_network_qos_policy(context, network_dic['id']))
        urlpath = self._strings_to_url(sdn_const.NETWORK_PATH,
                                       network_dic['id'])
        self._send_json_http_request(method=sdn_const.PUT,
                                     urlpath=urlpath,
                                     data=network_dic)

    @context_validator(sdn_const.NETWORK_PATH)
    @error_handler
    def delete_network_postcommit(self, context):
        network_dic = context._network
        network_dic[NETWORK_QOS_POLICY] = (
            self._get_network_qos_policy(context, network_dic['id']))
        urlpath = self._strings_to_url(sdn_const.NETWORK_PATH,
                                       network_dic['id'])
        self._send_json_http_request(method=sdn_const.DELETE,
                                     urlpath=urlpath,
                                     data=network_dic)

    @context_validator(sdn_const.PORT_PATH)
    @error_handler
    def update_port_postcommit(self, context):
        port_dic = context._port
        port_dic[NETWORK_QOS_POLICY] = (
            self._get_network_qos_policy(context, port_dic['network_id']))
        urlpath_port = self._strings_to_url(sdn_const.PORT_PATH,
                                            port_dic['id'])
        self._send_json_http_request(method=sdn_const.PUT,
                                     urlpath=urlpath_port,
                                     data=port_dic)

    @context_validator(sdn_const.PORT_PATH)
    @error_handler
    def delete_port_postcommit(self, context):
        port_dic = context._port
        port_dic[NETWORK_QOS_POLICY] = (
            self._get_network_qos_policy(context, port_dic['network_id']))
        urlpath_port = self._strings_to_url(sdn_const.PORT_PATH,
                                            port_dic['id'])
        self._send_json_http_request(method=sdn_const.DELETE,
                                     urlpath=urlpath_port,
                                     data=port_dic)

    @context_validator()
    @error_handler
    def bind_port(self, context):
        port_dic = context._port
        if self._is_send_bind_port(port_dic):
            port_dic[NETWORK_QOS_POLICY] = (
                self._get_network_qos_policy(context, port_dic['network_id']))
            self._send_json_http_request(method=sdn_const.POST,
                                         urlpath=sdn_const.PORT_PATH,
                                         data=port_dic)

    def _is_send_bind_port(self, port_context):
        """Verify that bind port is occur in compute context

        The request HTTP will occur only when the device owner is compute
        or dhcp.
        """
        device_owner = port_context['device_owner']
        return (device_owner and
                (device_owner.lower().startswith(
                 sdn_const.PORT_DEVICE_OWNER_COMPUTE) or
                 device_owner == neutron_const.DEVICE_OWNER_DHCP))

    def check_segment(self, segment):
        """Verify if a segment is valid for the SDN MechanismDriver.

        Verify if the requested segment is supported by SDN MD and return True
        or False to indicate this to callers.
        """
        network_type = segment[api.NETWORK_TYPE]
        return network_type in [constants.TYPE_VLAN]

    def check_segments(self, segments):
        """Verify if there is a segment in a list of segments that valid for
         the SDN MechanismDriver.

        Verify if the requested segments are supported by SDN MD and return
        True or False to indicate this to callers.
        """
        if segments:
            for segment in segments:
                if self.check_segment(segment):
                    return True
        return False

    def _send_json_http_request(self, urlpath, data, method):
        """send json http request to the SDN provider
        """
        dest_url = self._strings_to_url(self.url, self.domain, urlpath)

        data = jsonutils.dumps(data, indent=2)
        session = self._getSession()

        try:
            LOG.debug("Sending METHOD %(method)s URL %(url)s JSON %(data)s",
                      {'method': method, 'url': dest_url, 'data': data})
            r = session.request(method, url=dest_url,
                                headers=sdn_const.JSON_HTTP_HEADER,
                                data=data, timeout=self.timeout)
            LOG.debug("request status: %d", r.status_code)
            r.raise_for_status()
        except Exception as e:
            raise sdn_exc.SDNConnectionError(dest_url=dest_url, msg=e)

    def _get_network_qos_policy(self, context, net_id):
        return policy_object.QosPolicy.get_network_policy(
            context._plugin_context, net_id)



# Copyright 2015 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.agent.linux import dhcp
from neutron.extensions import extra_dhcp_opt as edo_ext


class DhcpOpt(object):
    def __init__(self, **kwargs):
        self.__dict__.update(ip_version=4)
        self.__dict__.update(kwargs)

    def __str__(self):
        return str(self.__dict__)


class MlnxDnsmasq(dhcp.Dnsmasq):
    _PREFIX = 'ff:00:00:00:00:00:02:00:00:02:c9:00:'
    _MIDDLE = ':00:00:'

    def _gen_client_id(self, port):
        mac_address = port.mac_address
        mac_first = mac_address[:8]
        mac_last = mac_address[9:]
        client_id = ''.join([self._PREFIX, mac_first, self._MIDDLE, mac_last])
        return client_id

    def _gen_client_id_opt(self, client_id):
        return DhcpOpt(opt_name=edo_ext.CLIENT_ID, opt_value=client_id)

    def _get_port_extra_dhcp_opts(self, port):
        client_id = self._gen_client_id(port)
        if hasattr(port, edo_ext.EXTRADHCPOPTS):
            for opt in port.extra_dhcp_opts:
                if opt.opt_name == edo_ext.CLIENT_ID:
                    opt.opt_value = client_id
                    return port.extra_dhcp_opts
            port.extra_dhcp_opts.append(self._gen_client_id_opt(client_id))
        else:
            setattr(port, edo_ext.EXTRADHCPOPTS,
                    [self._gen_client_id_opt(client_id)])
        return port.extra_dhcp_opts



# Copyright (c) 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.extensions import portbindings
from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2.drivers.mlnx import mech_mlnx
from neutron.tests.unit.plugins.ml2 import _test_mech_agent as base


class MlnxMechanismBaseTestCase(base.AgentMechanismBaseTestCase):
    VIF_TYPE = mech_mlnx.VIF_TYPE_IB_HOSTDEV
    CAP_PORT_FILTER = False
    AGENT_TYPE = mech_mlnx.AGENT_TYPE_MLNX
    VNIC_TYPE = portbindings.VNIC_DIRECT

    GOOD_MAPPINGS = {'fake_physical_network': 'fake_bridge'}
    GOOD_CONFIGS = {'interface_mappings': GOOD_MAPPINGS}

    BAD_MAPPINGS = {'wrong_physical_network': 'wrong_bridge'}
    BAD_CONFIGS = {'interface_mappings': BAD_MAPPINGS}

    AGENTS = [{'alive': True,
               'configurations': GOOD_CONFIGS,
               'host': 'host'}]
    AGENTS_DEAD = [{'alive': False,
                    'configurations': GOOD_CONFIGS,
                    'host': 'dead_host'}]
    AGENTS_BAD = [{'alive': False,
                   'configurations': GOOD_CONFIGS,
                   'host': 'bad_host_1'},
                  {'alive': True,
                   'configurations': BAD_CONFIGS,
                   'host': 'bad_host_2'}]

    def setUp(self):
        super(MlnxMechanismBaseTestCase, self).setUp()
        self.driver = mech_mlnx.MlnxMechanismDriver()
        self.driver.initialize()


class MlnxMechanismGenericTestCase(MlnxMechanismBaseTestCase,
                                   base.AgentMechanismGenericTestCase):
    pass


class MlnxMechanismLocalTestCase(MlnxMechanismBaseTestCase,
                                 base.AgentMechanismLocalTestCase):
    pass


class MlnxMechanismFlatTestCase(MlnxMechanismBaseTestCase,
                                base.AgentMechanismFlatTestCase):
    pass


class MlnxMechanismVlanTestCase(MlnxMechanismBaseTestCase,
                                base.AgentMechanismVlanTestCase):
    pass


class MlnxMechanismVifDetailsTestCase(MlnxMechanismBaseTestCase):
    def setUp(self):
        super(MlnxMechanismVifDetailsTestCase, self).setUp()

    def test_vif_details_contains_physical_net(self):
        VLAN_SEGMENTS = [{api.ID: 'vlan_segment_id',
                          api.NETWORK_TYPE: 'vlan',
                          api.PHYSICAL_NETWORK: 'fake_physical_network',
                          api.SEGMENTATION_ID: 1234}]

        context = base.FakePortContext(self.AGENT_TYPE,
                                       self.AGENTS,
                                       VLAN_SEGMENTS,
                                       portbindings.VNIC_DIRECT)
        segment = VLAN_SEGMENTS[0]
        agent = self.AGENTS[0]
        self.driver.try_to_bind_segment_for_agent(context, segment, agent)
        set({"physical_network": "fake_physical_network"}).issubset(
            set(context._bound_vif_details.items()))



# Copyright (c) 2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

import pep8

# Guidelines for writing new hacking checks
#
#  - Use only for Neutron specific tests. OpenStack general tests
#    should be submitted to the common 'hacking' module.
#  - Pick numbers in the range N3xx. Find the current test with
#    the highest allocated number and then pick the next value.
#  - Keep the test method code in the source file ordered based
#    on the N3xx value.
#  - List the new rule in the top level HACKING.rst file
#  - Add test cases for each new rule to
#    neutron/tests/unit/test_hacking.py

author_tag_re = re.compile(r"^\s*#\s*@?[aA]uthor|^\.\.\s+moduleauthor::")

_all_log_levels = {
    # NOTE(yamamoto): Following nova which uses _() for audit.
    'audit': '_',
    'error': '_LE',
    'info': '_LI',
    'warn': '_LW',
    'warning': '_LW',
    'critical': '_LC',
    'exception': '_LE',
}
_all_hints = set(_all_log_levels.values())


def _regex_for_level(level, hint):
    return r".*LOG\.%(level)s\(\s*((%(wrong_hints)s)\(|'|\")" % {
        'level': level,
        'wrong_hints': '|'.join(_all_hints - set([hint])),
    }


log_translation_hint = re.compile(
    '|'.join('(?:%s)' % _regex_for_level(level, hint)
             for level, hint in _all_log_levels.iteritems()))

oslo_namespace_imports_dot = re.compile(r"from[\s]*oslo[.]")
oslo_namespace_imports_root = re.compile(r"from[\s]*oslo[\s]*import[\s]*")


def validate_log_translations(logical_line, physical_line, filename):
    # Translations are not required in the test directory
    if "neutron/tests" in filename:
        return
    if pep8.noqa(physical_line):
        return

    msg = "N320: Log messages require translation hints!"
    if log_translation_hint.match(logical_line):
        yield (0, msg)


def check_oslo_namespace_imports(logical_line, blank_before, filename):
    if re.match(oslo_namespace_imports_dot, logical_line):
        msg = ("N323: '%s' must be used instead of '%s'.") % (
               logical_line.replace('oslo.', 'oslo_'),
               logical_line)
        yield(0, msg)
    elif re.match(oslo_namespace_imports_root, logical_line):
        msg = ("N323: '%s' must be used instead of '%s'.") % (
               logical_line.replace('from oslo import ', 'import oslo_'),
               logical_line)
        yield(0, msg)


def use_jsonutils(logical_line, filename):
    msg = "N321: jsonutils.%(fun)s must be used instead of json.%(fun)s"

    # Some files in the tree are not meant to be run from inside Neutron
    # itself, so we should not complain about them not using jsonutils
    json_check_skipped_patterns = [
        "neutron/plugins/openvswitch/agent/xenapi/etc/xapi.d/plugins/netwrap",
    ]

    for pattern in json_check_skipped_patterns:
        if pattern in filename:
            return

    if "json." in logical_line:
        json_funcs = ['dumps(', 'dump(', 'loads(', 'load(']
        for f in json_funcs:
            pos = logical_line.find('json.%s' % f)
            if pos != -1:
                yield (pos, msg % {'fun': f[:-1]})


def no_author_tags(physical_line):
    if author_tag_re.match(physical_line):
        physical_line = physical_line.lower()
        pos = physical_line.find('moduleauthor')
        if pos < 0:
            pos = physical_line.find('author')
        return pos, "N322: Don't use author tags"


def no_translate_debug_logs(logical_line, filename):
    """Check for 'LOG.debug(_(' and 'LOG.debug(_Lx('

    As per our translation policy,
    https://wiki.openstack.org/wiki/LoggingStandards#Log_Translation
    we shouldn't translate debug level logs.

    * This check assumes that 'LOG' is a logger.
    N319
    """
    for hint in _all_hints:
        if logical_line.startswith("LOG.debug(%s(" % hint):
            yield(0, "N319 Don't translate debug level logs")


def check_assert_called_once_with(logical_line, filename):
    # Try to detect unintended calls of nonexistent mock methods like:
    #    assert_called_once
    #    assertCalledOnceWith
    if 'neutron/tests/' in filename:
        if '.assert_called_once_with(' in logical_line:
            return
        if '.assertcalledonce' in logical_line.lower().replace('_', ''):
            msg = ("N323: Possible use of no-op mock method. "
                   "please use assert_called_once_with.")
            yield (0, msg)


def factory(register):
    register(validate_log_translations)
    register(use_jsonutils)
    register(no_author_tags)
    register(check_assert_called_once_with)
    register(no_translate_debug_logs)



# Copyright 2014 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib

import mock
from oslo_config import cfg
import testtools

from networking_mlnx.plugins.ml2.drivers.mlnx.agent import (
    mlnx_eswitch_neutron_agent)
from networking_mlnx.plugins.ml2.drivers.mlnx.agent import exceptions
from networking_mlnx.plugins.ml2.drivers.mlnx.agent import utils
from neutron.tests import base


class TestEswichManager(base.BaseTestCase):

    def setUp(self):
        super(TestEswichManager, self).setUp()

        class MockEswitchUtils(object):
            def __init__(self, endpoint, timeout):
                pass

        mock.patch('networking_mlnx.plugins.mlnx.agent.utils.EswitchManager',
                   new=MockEswitchUtils)

        with mock.patch.object(utils, 'zmq'):
            self.manager = mlnx_eswitch_neutron_agent.EswitchManager(
                {}, None, None)

    def test_get_not_exist_port_id(self):
        with testtools.ExpectedException(exceptions.MlnxException):
            self.manager.get_port_id_by_mac('no-such-mac')


class TestMlnxEswitchRpcCallbacks(base.BaseTestCase):

    def setUp(self):
        super(TestMlnxEswitchRpcCallbacks, self).setUp()
        agent = mock.Mock()
        self.rpc_callbacks = \
            mlnx_eswitch_neutron_agent.MlnxEswitchRpcCallbacks(
                'context',
                agent,
                agent)

    def test_port_update(self):
        port = {'mac_address': '10:20:30:40:50:60'}
        add_port_update = self.rpc_callbacks.agent.add_port_update
        self.rpc_callbacks.port_update('context', port=port)
        add_port_update.assert_called_once_with(port['mac_address'])


class TestEswitchAgent(base.BaseTestCase):

    def setUp(self):
        super(TestEswitchAgent, self).setUp()
        cfg.CONF.set_default('firewall_driver',
                             'neutron.agent.firewall.NoopFirewallDriver',
                             group='SECURITYGROUP')

        class MockFixedIntervalLoopingCall(object):
            def __init__(self, f):
                self.f = f

            def start(self, interval=0):
                self.f()

        mock.patch('neutron.openstack.common.loopingcall.'
                   'FixedIntervalLoopingCall',
                   new=MockFixedIntervalLoopingCall)

        with mock.patch.object(utils, 'zmq'):
            self.agent = mlnx_eswitch_neutron_agent.MlnxEswitchNeutronAgent(
                {})
        self.agent.plugin_rpc = mock.Mock()
        self.agent.context = mock.Mock()
        self.agent.agent_id = mock.Mock()
        self.agent.eswitch = mock.Mock()
        self.agent.eswitch.get_vnics_mac.return_value = []

    def test_treat_devices_added_returns_true_for_missing_device(self):
        attrs = {'get_devices_details_list.side_effect': Exception()}
        self.agent.plugin_rpc.configure_mock(**attrs)
        with contextlib.nested(
            mock.patch('networking_mlnx.plugins.ml2.drivers.mlnx.agent.'
                       'mlnx_eswitch_neutron_agent.EswitchManager.'
                       'get_vnics_mac',
                       return_value=[])):
            self.assertTrue(self.agent.treat_devices_added_or_updated([{}]))

    def _mock_treat_devices_added_updated(self, details, func_name):
        """Mock treat devices added.

        :param details: the details to return for the device
        :param func_name: the function that should be called
        :returns: whether the named function was called
        """
        with contextlib.nested(
            mock.patch('networking_mlnx.plugins.ml2.drivers.mlnx.agent.'
                       'mlnx_eswitch_neutron_agent.EswitchManager.'
                       'get_vnics_mac',
                       return_value=[]),
            mock.patch.object(self.agent.plugin_rpc,
                              'get_devices_details_list',
                              return_value=[details]),
            mock.patch.object(self.agent.plugin_rpc, 'update_device_up'),
            mock.patch.object(self.agent, func_name)
        ) as (vnics_fn, get_dev_fn, upd_dev_up, func):
            self.assertFalse(self.agent.treat_devices_added_or_updated([{}]))
        return (func.called, upd_dev_up.called)

    def test_treat_devices_added_updates_known_port(self):
        details = mock.MagicMock()
        details.__contains__.side_effect = lambda x: True
        func, dev_up = self._mock_treat_devices_added_updated(details,
                                                              'treat_vif_port')
        self.assertTrue(func)
        self.assertTrue(dev_up)

    def test_treat_devices_added_updates_known_port_admin_down(self):
        details = {'port_id': '1234567890',
                   'device': '01:02:03:04:05:06',
                   'network_id': '123456789',
                   'network_type': 'vlan',
                   'physical_network': 'default',
                   'segmentation_id': 2,
                   'admin_state_up': False}
        func, dev_up = self._mock_treat_devices_added_updated(details,
                                                              'treat_vif_port')
        self.assertTrue(func)
        self.assertFalse(dev_up)

    def test_treat_devices_removed_returns_true_for_missing_device(self):
        with mock.patch.object(self.agent.plugin_rpc, 'update_device_down',
                               side_effect=Exception()):
            self.assertTrue(self.agent.treat_devices_removed([{}]))

    def test_treat_devices_removed_releases_port(self):
        details = dict(exists=False)
        with mock.patch.object(self.agent.plugin_rpc, 'update_device_down',
                               return_value=details):
            with mock.patch.object(self.agent.eswitch,
                                   'port_release') as port_release:
                self.assertFalse(self.agent.treat_devices_removed([{}]))
                self.assertTrue(port_release.called)

    def _test_process_network_ports(self, port_info):
        with contextlib.nested(
            mock.patch.object(self.agent, 'treat_devices_added_or_updated',
                              return_value=False),
            mock.patch.object(self.agent, 'treat_devices_removed',
                              return_value=False)
        ) as (device_added_updated, device_removed):
            self.assertFalse(self.agent.process_network_ports(port_info))
            device_added_updated.assert_called_once_with(
                port_info['added'] | port_info['updated'])
            device_removed.assert_called_once_with(port_info['removed'])

    def test_process_network_ports(self):
        self._test_process_network_ports(
            {'current': set(['10:20:30:40:50:60']),
             'updated': set(),
             'added': set(['11:21:31:41:51:61']),
             'removed': set(['13:23:33:43:53:63'])})

    def test_process_network_ports_with_updated_ports(self):
        self._test_process_network_ports(
            {'current': set(['10:20:30:40:50:60']),
             'updated': set(['12:22:32:42:52:62']),
             'added': set(['11:21:31:41:51:61']),
             'removed': set(['13:23:33:43:53:63'])})

    def test_add_port_update(self):
        mac_addr = '10:20:30:40:50:60'
        self.agent.add_port_update(mac_addr)
        self.assertEqual(set([mac_addr]), self.agent.updated_ports)

    def _mock_scan_ports(self, vif_port_set, previous,
                         updated_ports, sync=False):
        self.agent.updated_ports = updated_ports
        with mock.patch.object(self.agent.eswitch, 'get_vnics_mac',
                               return_value=vif_port_set):
            return self.agent.scan_ports(previous, sync)

    def test_scan_ports_return_current_for_unchanged_ports(self):
        vif_port_set = set([1, 2])
        previous = dict(current=set([1, 2]), added=set(),
                        removed=set(), updated=set())
        expected = dict(current=vif_port_set, added=set(),
                        removed=set(), updated=set())
        actual = self._mock_scan_ports(vif_port_set,
                                       previous, set())
        self.assertEqual(expected, actual)

    def test_scan_ports_return_port_changes(self):
        vif_port_set = set([1, 3])
        previous = dict(current=set([1, 2]), added=set(),
                        removed=set(), updated=set())
        expected = dict(current=vif_port_set, added=set([3]),
                        removed=set([2]), updated=set())
        actual = self._mock_scan_ports(vif_port_set,
                                       previous, set())
        self.assertEqual(expected, actual)

    def test_scan_ports_with_updated_ports(self):
        vif_port_set = set([1, 3, 4])
        previous = dict(current=set([1, 2, 4]), added=set(),
                        removed=set(), updated=set())
        expected = dict(current=vif_port_set, added=set([3]),
                        removed=set([2]), updated=set([4]))
        actual = self._mock_scan_ports(vif_port_set,
                                       previous, set([4]))
        self.assertEqual(expected, actual)

    def test_scan_ports_with_unknown_updated_ports(self):
        vif_port_set = set([1, 3, 4])
        previous = dict(current=set([1, 2, 4]), added=set(),
                        removed=set(), updated=set())
        expected = dict(current=vif_port_set, added=set([3]),
                        removed=set([2]), updated=set([4]))
        actual = self._mock_scan_ports(vif_port_set,
                                       previous,
                                       updated_ports=set([4, 5]))
        self.assertEqual(expected, actual)



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import socket
import time

import eventlet
eventlet.monkey_patch()

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging
from oslo_service import loopingcall

from neutron.agent import rpc as agent_rpc
from neutron.agent import securitygroups_rpc as sg_rpc
from neutron.common import constants as q_constants
from neutron.common import topics
from neutron import context
from neutron.i18n import _LE, _LI, _LW
from neutron.plugins.common import constants as p_const
from neutron.plugins.ml2.drivers.mlnx.agent import config  # noqa
from neutron.plugins.ml2.drivers.mlnx import mech_mlnx

from networking_mlnx.plugins.ml2.drivers.mlnx.agent import exceptions
from networking_mlnx.plugins.ml2.drivers.mlnx.agent import utils


LOG = logging.getLogger(__name__)


class EswitchManager(object):
    def __init__(self, interface_mappings, endpoint, timeout):
        self.utils = utils.EswitchUtils(endpoint, timeout)
        self.interface_mappings = interface_mappings
        self.network_map = {}
        self.utils.define_fabric_mappings(interface_mappings)

    def get_port_id_by_mac(self, port_mac):
        for network_id, data in self.network_map.iteritems():
            for port in data['ports']:
                if port['port_mac'] == port_mac:
                    return port['port_id']
        LOG.error(_LE("Agent cache inconsistency - port id "
                      "is not stored for %s"), port_mac)
        raise exceptions.MlnxException(err_msg=("Agent cache inconsistency, "
                                                "check logs"))

    def get_vnics_mac(self):
        return set(self.utils.get_attached_vnics().keys())

    def vnic_port_exists(self, port_mac):
        return port_mac in self.utils.get_attached_vnics()

    def remove_network(self, network_id):
        if network_id in self.network_map:
            del self.network_map[network_id]
        else:
            LOG.debug("Network %s not defined on Agent.", network_id)

    def port_down(self, network_id, physical_network, port_mac):
        """Sets port to down.

        Check internal network map for port data.
        If port exists set port to Down
        """
        for network_id, data in self.network_map.iteritems():
            for port in data['ports']:
                if port['port_mac'] == port_mac:
                    self.utils.port_down(physical_network, port_mac)
                    return
        LOG.info(_LI('Network %s is not available on this agent'), network_id)

    def port_up(self, network_id, network_type,
                physical_network, seg_id, port_id, port_mac):
        """Sets port to up.

        Update internal network map with port data.
        - Check if vnic defined
        - configure eswitch vport
        - set port to Up
        """
        LOG.debug("Connecting port %s", port_id)

        if network_id not in self.network_map:
            self.provision_network(port_id, port_mac,
                                   network_id, network_type,
                                   physical_network, seg_id)
        net_map = self.network_map[network_id]
        net_map['ports'].append({'port_id': port_id, 'port_mac': port_mac})

        if network_type == p_const.TYPE_VLAN:
            LOG.info(_LI('Binding Segmentation ID %(seg_id)s '
                         'to eSwitch for vNIC mac_address %(mac)s'),
                     {'seg_id': seg_id,
                      'mac': port_mac})
            self.utils.set_port_vlan_id(physical_network,
                                        seg_id,
                                        port_mac)
            self.utils.port_up(physical_network, port_mac)
        else:
            LOG.error(_LE('Unsupported network type %s'), network_type)

    def port_release(self, port_mac):
        """Clear port configuration from eSwitch."""
        for network_id, net_data in self.network_map.iteritems():
            for port in net_data['ports']:
                if port['port_mac'] == port_mac:
                    self.utils.port_release(net_data['physical_network'],
                                            port['port_mac'])
                    return
        LOG.info(_LI('Port_mac %s is not available on this agent'), port_mac)

    def provision_network(self, port_id, port_mac,
                          network_id, network_type,
                          physical_network, segmentation_id):
        LOG.info(_LI("Provisioning network %s"), network_id)
        if network_type == p_const.TYPE_VLAN:
            LOG.debug("Creating VLAN Network")
        else:
            LOG.error(_LE("Unknown network type %(network_type)s "
                          "for network %(network_id)s"),
                      {'network_type': network_type,
                       'network_id': network_id})
            return
        data = {
            'physical_network': physical_network,
            'network_type': network_type,
            'ports': [],
            'vlan_id': segmentation_id}
        self.network_map[network_id] = data


class MlnxEswitchRpcCallbacks(sg_rpc.SecurityGroupAgentRpcCallbackMixin):

    # Set RPC API version to 1.0 by default.
    # history
    #   1.1 Support Security Group RPC
    target = oslo_messaging.Target(version='1.1')

    def __init__(self, context, agent, sg_agent):
        super(MlnxEswitchRpcCallbacks, self).__init__()
        self.context = context
        self.agent = agent
        self.eswitch = agent.eswitch
        self.sg_agent = sg_agent

    def network_delete(self, context, **kwargs):
        LOG.debug("network_delete received")
        network_id = kwargs.get('network_id')
        if not network_id:
            LOG.warning(_LW("Invalid Network ID, cannot remove Network"))
        else:
            LOG.debug("Delete network %s", network_id)
            self.eswitch.remove_network(network_id)

    def port_update(self, context, **kwargs):
        port = kwargs.get('port')
        self.agent.add_port_update(port['mac_address'])
        LOG.debug("port_update message processed for port with mac %s",
                  port['mac_address'])


class MlnxEswitchNeutronAgent(object):

    def __init__(self, interface_mapping):
        self._polling_interval = cfg.CONF.AGENT.polling_interval
        self._setup_eswitches(interface_mapping)
        configurations = {'interface_mappings': interface_mapping}
        self.agent_state = {
            'binary': 'neutron-mlnx-agent',
            'host': cfg.CONF.host,
            'topic': q_constants.L2_AGENT_TOPIC,
            'configurations': configurations,
            'agent_type': mech_mlnx.AGENT_TYPE_MLNX,
            'start_flag': True}
        # Stores port update notifications for processing in main rpc loop
        self.updated_ports = set()
        self.context = context.get_admin_context_without_session()
        self.plugin_rpc = agent_rpc.PluginApi(topics.PLUGIN)
        self.sg_plugin_rpc = sg_rpc.SecurityGroupServerRpcApi(topics.PLUGIN)
        self.sg_agent = sg_rpc.SecurityGroupAgentRpc(self.context,
                self.sg_plugin_rpc)
        self._setup_rpc()

    def _setup_eswitches(self, interface_mapping):
        daemon = cfg.CONF.ESWITCH.daemon_endpoint
        timeout = cfg.CONF.ESWITCH.request_timeout
        self.eswitch = EswitchManager(interface_mapping, daemon, timeout)

    def _report_state(self):
        try:
            devices = len(self.eswitch.get_vnics_mac())
            self.agent_state.get('configurations')['devices'] = devices
            self.state_rpc.report_state(self.context,
                                        self.agent_state)
            self.agent_state.pop('start_flag', None)
        except Exception:
            LOG.exception(_LE("Failed reporting state!"))

    def _setup_rpc(self):
        self.agent_id = 'mlnx-agent.%s' % socket.gethostname()
        LOG.info(_LI("RPC agent_id: %s"), self.agent_id)

        self.topic = topics.AGENT
        self.state_rpc = agent_rpc.PluginReportStateAPI(topics.PLUGIN)
        # RPC network init
        # Handle updates from service
        self.endpoints = [MlnxEswitchRpcCallbacks(self.context, self,
                                                  self.sg_agent)]
        # Define the listening consumers for the agent
        consumers = [[topics.PORT, topics.UPDATE],
                     [topics.NETWORK, topics.DELETE],
                     [topics.SECURITY_GROUP, topics.UPDATE]]
        self.connection = agent_rpc.create_consumers(self.endpoints,
                                                     self.topic,
                                                     consumers)

        report_interval = cfg.CONF.AGENT.report_interval
        if report_interval:
            heartbeat = loopingcall.FixedIntervalLoopingCall(
                self._report_state)
            heartbeat.start(interval=report_interval)

    def add_port_update(self, port):
        self.updated_ports.add(port)

    def scan_ports(self, previous, sync):
        cur_ports = self.eswitch.get_vnics_mac()
        port_info = {'current': cur_ports}
        updated_ports = self.updated_ports
        self.updated_ports = set()
        if sync:
            # Either it's the first iteration or previous iteration had
            # problems.
            port_info['added'] = cur_ports
            port_info['removed'] = ((previous['removed'] | previous['current'])
                                    - cur_ports)
            port_info['updated'] = ((previous['updated'] | updated_ports)
                                    & cur_ports)
        else:
            # Shouldn't process updates for not existing ports
            port_info['added'] = cur_ports - previous['current']
            port_info['removed'] = previous['current'] - cur_ports
            port_info['updated'] = updated_ports & cur_ports
        return port_info

    def process_network_ports(self, port_info):
        resync_a = False
        resync_b = False
        device_added_updated = port_info['added'] | port_info['updated']

        if device_added_updated:
            resync_a = self.treat_devices_added_or_updated(
                device_added_updated)
        if port_info['removed']:
            resync_b = self.treat_devices_removed(port_info['removed'])
        # If one of the above opertaions fails => resync with plugin
        return (resync_a | resync_b)

    def treat_vif_port(self, port_id, port_mac,
                       network_id, network_type,
                       physical_network, segmentation_id,
                       admin_state_up):
        if self.eswitch.vnic_port_exists(port_mac):
            if admin_state_up:
                self.eswitch.port_up(network_id,
                                     network_type,
                                     physical_network,
                                     segmentation_id,
                                     port_id,
                                     port_mac)
            else:
                self.eswitch.port_down(network_id, physical_network, port_mac)
        else:
            LOG.debug("No port %s defined on agent.", port_id)

    def treat_devices_added_or_updated(self, devices):
        try:
            devs_details_list = self.plugin_rpc.get_devices_details_list(
                self.context,
                devices,
                self.agent_id)
        except Exception as e:
            LOG.debug("Unable to get device details for devices "
                      "with MAC address %(devices)s: due to %(exc)s",
                      {'devices': devices, 'exc': e})
            # resync is needed
            return True

        for dev_details in devs_details_list:
            device = dev_details['device']
            LOG.info(_LI("Adding or updating port with mac %s"), device)

            if 'port_id' in dev_details:
                LOG.info(_LI("Port %s updated"), device)
                LOG.debug("Device details %s", str(dev_details))
                self.treat_vif_port(dev_details['port_id'],
                                    dev_details['device'],
                                    dev_details['network_id'],
                                    dev_details['network_type'],
                                    dev_details['physical_network'],
                                    dev_details['segmentation_id'],
                                    dev_details['admin_state_up'])
                if dev_details.get('admin_state_up'):
                    LOG.debug("Setting status for %s to UP", device)
                    self.plugin_rpc.update_device_up(
                        self.context, device, self.agent_id)
                else:
                    LOG.debug("Setting status for %s to DOWN", device)
                    self.plugin_rpc.update_device_down(
                        self.context, device, self.agent_id)
            else:
                LOG.debug("Device with mac_address %s not defined "
                          "on Neutron Plugin", device)
        return False

    def treat_devices_removed(self, devices):
        resync = False
        for device in devices:
            LOG.info(_LI("Removing device with mac_address %s"), device)
            try:
                port_id = self.eswitch.get_port_id_by_mac(device)
                dev_details = self.plugin_rpc.update_device_down(self.context,
                                                                 port_id,
                                                                 self.agent_id,
                                                                 cfg.CONF.host)
            except Exception as e:
                LOG.debug("Removing port failed for device %(device)s "
                          "due to %(exc)s", {'device': device, 'exc': e})
                resync = True
                continue
            if dev_details['exists']:
                LOG.info(_LI("Port %s updated."), device)
            else:
                LOG.debug("Device %s not defined on plugin", device)
            self.eswitch.port_release(device)
        return resync

    def _port_info_has_changes(self, port_info):
        return (port_info['added'] or
                port_info['removed'] or
                port_info['updated'])

    def run(self):
        LOG.info(_LI("eSwitch Agent Started!"))
        sync = True
        port_info = {'current': set(),
                     'added': set(),
                     'removed': set(),
                     'updated': set()}
        while True:
            start = time.time()
            try:
                port_info = self.scan_ports(previous=port_info, sync=sync)
            except exceptions.RequestTimeout:
                LOG.exception(_LE("Request timeout in agent event loop "
                                  "eSwitchD is not responding - exiting..."))
                sync = True
                continue
            if sync:
                LOG.info(_LI("Agent out of sync with plugin!"))
                sync = False
            if self._port_info_has_changes(port_info):
                LOG.debug("Starting to process devices in:%s", port_info)
                try:
                    sync = self.process_network_ports(port_info)
                except Exception:
                    LOG.exception(_LE("Error in agent event loop"))
                    sync = True
            # sleep till end of polling interval
            elapsed = (time.time() - start)
            if (elapsed < self._polling_interval):
                time.sleep(self._polling_interval - elapsed)
            else:
                LOG.debug("Loop iteration exceeded interval "
                          "(%(polling_interval)s vs. %(elapsed)s)",
                          {'polling_interval': self._polling_interval,
                           'elapsed': elapsed})
